#!/usr/bin/env perl

#This is an extended version of:
# Get status of Linux software RAID for SNMP / Nagios
# Author: Michal Ludvig <michal@logix.cz>
#         http://www.logix.cz/michal/devel/nagios
#
#It's intended to work with Synology DiskStation
#Original found here: http://www.logix.cz/michal/devel/nagios/nagios-linux-swraid.pl
#
#How a physical disk maps to a dev and volume found here:
#http://forum.synology.com/wiki/index.php/How_to_give_your_disks_and_files_a_health_check_and_repair_basic_faults
# Author: Diego d'Ambra (da@panorama9.com)

#includes
use strict;
use Getopt::Long;

my $file   = "/proc/mdstat";
my $device = "all";

#get command line options.
GetOptions(
    'file=s'   => \$file,
    'device=s' => \$device,
    'help'     => sub { &usage() }
);

#strip leading "/dev/" from --device in case it has been given
$device =~ s/^\/dev\///;

#return codes for Nagios
my %ERRORS = ( 'OK' => 0, 'WARNING' => 1, 'CRITICAL' => 2, 'UNKNOWN' => 3, 'DEPENDENT' => 4 );

#this is a global return value - set to the worst result we get overall
my $retval = 0;

#info about physical disks
my %hdd_info;
my $hhd_cmd = _locate_tool('/usr/syno/bin/syno_hdd_util');
if ( defined($hhd_cmd) ) {

    #get info about hdd
    open( CMD, "$hhd_cmd --ssd_detect |" );
    my $l;
    while (<CMD>) {
        chomp;
        next if ( !/^([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+(\/dev\/\wd\w)\s+([^\s]+)/ );
        $hdd_info{$4} = {
            'model'    => $1,
            'firmware' => $2,
            'serial'   => $3,
            'is_ssd'   => $5,
        };
    }
    close(CMD);
}

#info about model, serial, hostname, firmware, number of disks and uptime
my $model    = '/proc/sys/kernel/syno_hw_version';
my $serial   = '/proc/sys/kernel/syno_serial';
my $hostname = '/proc/sys/kernel/hostname';
my $firmware = "$1.$2-$3"
  if ( _read_file('/etc.defaults/VERSION') =~ /majorversion="(\d+)"minorversion="(\d+)".+buildnumber="(\d+)/ );
my $uptime = $1 if ( _read_file('/proc/uptime') =~ /^([^\.]+)/ );
print "OK - SYSTEM, model: "
  . _read_file($model)
  . ", ser: "
  . _read_file($serial)
  . ", hn: "
  . _read_file($hostname)
  . ", fw: $firmware, disks: "
  . scalar( keys %hdd_info )
  . ", up: $uptime\n";

#map logical volumes to devices
my (%dev_info);
my $df_cmd = _locate_tool('df');
if ( defined($df_cmd) ) {

    #size and free space of each logical volume
    open( CMD, "$df_cmd |" );
    my $l;
    while (<CMD>) {
        chomp;
        $l = $_;

        #df may split filesystem status into 2 lines and Busybox doesn't support cmd-line option -P (join if needed)
        $l .= ' ' . <CMD> if ( $l =~ /^[^\s]+$/ );

        #split line
        if ( $l =~ /^([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+[^\s]+\s+([^\s]+)/ ) {
            my $fs        = $1;
            my $size      = $2;
            my $used      = $3;
            my $available = $4;
            my $mount     = $5;
            my $vol_num   = $1 if ( $mount =~ /(\d+)$/ );
            next unless ($vol_num);

            #single or multi bay NAS?
            my $dev;
            if ( -e '/dev/md' . ( $vol_num + 1 ) ) {
                $dev = '/dev/md' . ( $vol_num + 1 );
            }
            elsif ( -e '/dev/hd' . ( $vol_num + 2 ) ) {
                $dev = '/dev/hd' . ( $vol_num + 2 );
            }

            #store device info (only map first found)
            if ( !exists( $dev_info{$dev} ) ) {
                $dev_info{$dev} = {
                    'available' => $available,
                    'size'      => $size,
                    'used'      => $used,
                    'mount'     => $mount,
                };
            }
        }
    }
    close(CMD);

}

#open file (mdstat)
my ( %active_devs, %failed_devs, %spare_devs, %drives );
open FILE, "< $file" or die "Can't open $file : $!";
while (<FILE>) {
    next if !/^(md\d+)+\s*:\s*\w+\s*(raid\d+)/;
    next if $device ne "all" and $device ne $1;
    my $dev  = $1;
    my $raid = $2;

    my @array = split(/ /);
    for $_ (@array) {
        next if !/(\w+)\[\d+\](\(.\))*/;
        if ( $2 eq "(F)" ) {
            $failed_devs{$dev} .= "$1,";
        }
        elsif ( $2 eq "(S)" ) {
            $spare_devs{$dev} .= "$1,";
        }
        else {
            $active_devs{$dev} .= "$1,";
        }

        #remember drive, since we need detailed info about each physical disk
        my $ph_disk = $1;
        $ph_disk =~ s/\d+$//;
        $drives{$ph_disk} = 1;

    }
    if ( !defined( $active_devs{$dev} ) ) {
        $active_devs{$dev} = "none";
    }
    else {
        $active_devs{$dev} =~ s/,$//;
    }
    if ( !defined( $spare_devs{$dev} ) ) {
        $spare_devs{$dev} = "none";
    }
    else {
        $spare_devs{$dev} =~ s/,$//;
    }
    if ( !defined( $failed_devs{$dev} ) ) {
        $failed_devs{$dev} = "none";
    }
    else {
        $failed_devs{$dev} =~ s/,$//;

    }

    $_ = <FILE>;
    /\[(\d+)\/(\d+)\]\s+\[(.*)\]$/;
    my $devs_total = $1;
    my $devs_up    = $2;
    my $stat       = $3;
    my $result     = "OK";
    if ( $devs_total > $devs_up or $failed_devs{$dev} ne "none" ) {
        $result = "CRITICAL";
        $retval = $ERRORS{"CRITICAL"};
    }

    #print result
    print "$result - RAID, $dev $raid [$stat] $devs_up/$devs_total "
      . "(A:$active_devs{$dev} F:$failed_devs{$dev} S:$spare_devs{$dev})"
      . (
        exists( $dev_info{ '/dev/' . $dev } )
        ? ", size: $dev_info{'/dev/'.$dev}{'size'}, free: $dev_info{'/dev/'.$dev}{'available'}, name: $dev_info{'/dev/'.$dev}{'mount'}"
        : ''
      ) . "\n";
}
close FILE;

#health of each physical disk
my $smartctl_cmd = _locate_tool('smartctl');
if ( defined("$smartctl_cmd") ) {
    foreach my $d ( sort keys %drives ) {

        #map found drive to physical disk
        my $drive = '/dev/' . $d;

        #get drive info
        my ( $model, $size );
        open( CMD, "$smartctl_cmd -d ata -i $drive |" ) || next;
        while (<CMD>) {
            $model = $1 if (/^Device Model:\s+(.+)/);
            $size  = $1 if (/^User Capacity:\s+([\d\.,]+)/);
        }
        close(CMD);
        next unless ( defined($model) );
        $size =~ s/(,|\.)//g;

        #get smartctrl status
        my $health = 'CRITICAL';
        open( CMD, "$smartctl_cmd -d ata -H $drive |" ) || next;
        while (<CMD>) {
            $health = 'OK' if (/test result:\s+passed/i);
        }
        close(CMD);

        #get disk temp
        my $temp = -1;
        open( CMD, "$smartctl_cmd -d ata -A $drive |" ) || next;
        while (<CMD>) {
            $temp = $1 if (/Temperature_Celsius.+?(\d+)$/);
        }
        close(CMD);

        #print result
        print "$health - HD, model: $model, size: $size, serial: "
          . ( exists( $hdd_info{$drive} ) ? $hdd_info{$drive}{'serial'} : 'unknown' )
          . ", temp: $temp, dev: $d\n";
        $retval = $ERRORS{"CRITICAL"} unless ( $health eq 'OK' );

    }
}

#done
exit $retval;

# =====
sub _locate_tool {
    my $tool = shift;

    #find tool
    return $1 if ( `which $tool` =~ /(.+)/ );
}

sub _read_file {
    my $file = shift;

    my $ret_file;
    open( FILE, "<$file" ) || return "unknown";
    while (<FILE>) {
        chomp;
        $ret_file .= $_;
    }
    close(FILE);

    #return content of file
    return $ret_file;
}

sub usage() {
    printf( "
Check status of Linux SW RAID

This is an extended version of the script created by:

Author: Michal Ludvig <michal\@logix.cz> (c) 2006
        http://www.logix.cz/michal/devel/nagios
        
Continued version by:        
        Diego d'Ambra <da\@panorama9.com>
        http://panorama9.com

Usage: disk_info.pl [options]

  --file=<filename>    Name of file to parse. Default is /proc/mdstat
  --device=<device>    Name of MD device, e.g. md0. Default is \"all\"

" );
    exit(1);
}
